import os
from typing import Dict, Optional
import raichu.data
from dataclasses import dataclass

SMILES_FILE = os.path.join(os.path.dirname(raichu.data.__file__), 'smiles.txt')
METADATA_FILE = os.path.join(os.path.dirname(raichu.data.__file__), 'metadata.txt')


@dataclass
class ParasSubstrate:
    name: str = ''
    smiles: str = ''
    abbreviation: str = ''
    type: str = ''
    chirality: Optional[str] = None
    norine: Optional[str] = None

    @classmethod
    def from_line(cls, line):
        line_data = line.split('\t')
        for i in range(len(line_data)):
            line_data[i] = line_data[i].strip()
            if not line_data[i] or line_data[i] == 'None':
                line_data[i] = None
        name, smiles, abbreviation, substrate_type, chirality, norine = line_data
        assert name and smiles and abbreviation and substrate_type
        if not norine:
            norine = 'X'

        return cls(name, smiles, abbreviation, substrate_type, chirality, norine)


def load_smiles() -> Dict[str, str]:
    name_to_smiles = {}
    with open(SMILES_FILE, 'r') as smiles_file:
        smiles_file.readline()
        for line in smiles_file:
            name, smiles = line.strip().split('\t')
            name_to_smiles[name] = smiles
    return name_to_smiles


def load_metadata() -> Dict[str, ParasSubstrate]:
    name_to_metadata = {}
    with open(METADATA_FILE, 'r') as metadata:
        metadata.readline()
        for line in metadata:
            substrate = ParasSubstrate.from_line(line)
            name_to_metadata[substrate.name] = substrate
    return name_to_metadata


UNKNOWN_SUBSTRATE = "**Unknown**"

_SMILES = load_smiles()
_SMILES[UNKNOWN_SUBSTRATE] = "NC(*)C(=O)O"
_METADATA = load_metadata()
_METADATA[UNKNOWN_SUBSTRATE] = ParasSubstrate(UNKNOWN_SUBSTRATE, _SMILES[UNKNOWN_SUBSTRATE], "X", "amino_acid",
                                              "L", "X")


def get_smiles(substrate_name: str) -> str:
    """
    Return SMILES string from substrate name

    substrate_name: str. Unknown substrate: use UNKNOWN_SUBSTRATE
    """
    if not isinstance(substrate_name, str):
        raise TypeError(f"Expected str, got {type(substrate_name)}.")

    if substrate_name in _SMILES:
        return _SMILES[substrate_name]

    raise ValueError(f"Cannot fetch SMILES string for {substrate_name}.")


def get_metadata(substrate_name: str) -> ParasSubstrate:
    """
    Return PARAS metadata from substrate name

    substrate_name: str. Unknown substrate: use UNKNOWN_SUBSTRATE
    """
    if not isinstance(substrate_name, str):
        raise TypeError(f"Expected str, got {type(substrate_name)}.")

    if substrate_name in _METADATA:
        return _METADATA[substrate_name]

    raise ValueError(f"Cannot fetch metadata for {substrate_name}.")




from enum import Enum, unique


@unique
class DomainSuperClass(Enum):
    """
    An Enum representing the superclass of domains supported by RAIChU
    """
    RECOGNITION = 1
    SYNTHESIS = 2
    CARRIER = 3
    TAILORING = 4
    TERMINATION = 5

    @staticmethod
    def from_string(label: str) -> "DomainSuperClass":
        for value in DomainSuperClass:
            if str(value.name) == label:
                return value
        raise ValueError(f"Unknown domain superclass: {label}")


@unique
class RecognitionDomainType(Enum):
    """
    An Enum representing the types of recognition domains supported by RAIChU
    """
    A = 1
    AT = 2
    DUMMY_A = 3
    DUMMY_AT = 4
    CAL = 5

    @staticmethod
    def from_string(label: str) -> "RecognitionDomainType":
        for value in RecognitionDomainType:
            if str(value.name) == label:
                return value
        raise ValueError(f"Unknown recognition domain type: {label}")


@unique
class TerminationDomainType(Enum):
    """
    An Enum representing the types of termination domains supported by RAIChU
    """
    TE = 1
    TD = 2
    DUMMY_TE = 3
    DUMMY_TD = 4

    @staticmethod
    def from_string(label: str) -> "TerminationDomainType":
        for value in TerminationDomainType:
            if str(value.name) == label:
                return value
        raise ValueError(f"Unknown termination domain type: {label}")


@unique
class SynthesisDomainType(Enum):
    """
    An Enum representing the types of synthesis domains supported by RAIChU
    """
    C = 1
    KS = 2
    DUMMY_C = 3
    DUMMY_KS = 4

    @staticmethod
    def from_string(label: str) -> "SynthesisDomainType":
        for value in SynthesisDomainType:
            if str(value.name) == label:
                return value
        raise ValueError(f"Unknown synthesis domain type: {label}")


@unique
class CarrierDomainType(Enum):
    """
    An Enum representing the types of carrier domains supported by RAIChU
    """
    PCP = 1
    ACP = 2
    DUMMY_PCP = 3
    DUMMY_ACP = 4

    @staticmethod
    def from_string(label: str) -> "CarrierDomainType":
        for value in CarrierDomainType:
            if str(value.name) == label:
                return value
        raise ValueError(f"Unknown carrier domain type: {label}")


@unique
class TailoringDomainType(Enum):
    """
    An Enum representing the types of tailoring domains supported by RAIChU
    """
    ER = 1
    KR = 2
    DH = 3
    E = 5
    nMT = 6
    UNKNOWN = 7
    DUMMY_ALMT = 8 #Alpha-L-Methyltransferase
    DUMMY_AMT = 9 #Alpha-Methyltransferase
    DUMMY_SC = 10 #Smalles cyclase for creating pyran/furan rings
    DUMMY_ZDH = 11 #E-configured double bonds
    DUMMY_EDH = 12 #Z-configured double bonds
    DUMMY_AH = 13 #Alpha-hydroxylase
    DUMMY_GDH = 14 #Gamma-beta-dehydrogenase
    DUMMY_ZGDH = 15 #Z-Gamma-beta-dehydrogenase
    DUMMY_EGDH = 16 #E-Gamma-beta-dehydrogenase
    DUMMY_OMT = 17 #Beta-Hydroxymethyltransferase
    DUMMY_BMT = 18 #Beta-Methyltransferase
    DUMMY_ER = 19
    DUMMY_KR = 20
    DUMMY_DH = 21
    DUMMY_E = 22
    DUMMY_nMT = 23
    CYC = 24
    OX = 25
    DUMMY_EMO = 26

    @staticmethod
    def from_string(label: str) -> "TailoringDomainType":
        for value in TailoringDomainType:
            if str(value.name) == label:
                return value
        raise ValueError(f"Unknown tailoring domain type: {label}")


@unique
class KRDomainSubtype(Enum):
    """
    An Enum representing the subtypes of KR domain supported by RAIChU
    """
    A1 = 1
    A2 = 2
    B1 = 3
    B2 = 4
    C1 = 5
    C2 = 6
    UNKNOWN = 7

    @staticmethod
    def from_string(label: str) -> "KRDomainSubtype":
        for value in KRDomainSubtype:
            if str(value.name) == label:
                return value
        raise ValueError(f"Unknown KR domain subtype: {label}")


@unique
class KSDomainSubtype(Enum):
    """
    An Enum representing the subtypes of KS domain supported by RAIChU
    """
    CIS = 1
    UNKNOWN = 2
    NON_ELONGATING_BETA_L_OH = 3
    BETA_OH = 4
    NON_ELONGATING_ALPHAME_EDB = 5
    OXI = 6
    ST = 7
    BETA_OH_EDB = 8
    UNST = 9
    BETA_D_OH = 10
    NON_ELONGATING_BETA_OH = 11
    LACST = 12
    OUT = 13
    SHDB = 14
    BETA_OH_KETO = 15
    OXA = 16
    ACST = 17
    ALPHAME_BETAOH = 18
    NON_ELONGATING_OXA = 19
    BETA_D_OME = 20
    ARST = 21
    ALPHABETA_OH = 22
    RED_SHDB = 23
    BR = 24
    BETA_ME = 25
    EDB = 26
    BETA_MEDB = 27
    NON_ELONGATING_DB = 28
    ALPHAME_EDB = 29
    ALPHAME = 30
    ALPHAME_BETA_L_OH = 31
    AA = 32
    DB = 33
    RED = 34
    PYR = 35
    ALPHAME_BETA_D_OH = 36
    NON_ELONGATING = 37
    MEOST = 38
    BETA_L_OH = 39
    ZDB = 40
    KETO = 41
    ALPHA_OH = 42
    NON_ELONGATING_PYR = 43
    MISCELLANEOUS = 44
    EXOMETHYLENE = 45
    ALPHAME_ZDB = 46
    ALPHA_D_ME_SHDB = 47
    ALPHAME_DB = 48
    ALPHAME_KETO = 49
    NON_ELONGATING_SHDB = 50

    @staticmethod
    def from_string(label: str) -> "KSDomainSubtype":
        for value in KSDomainSubtype:
            if str(value.name) == label:
                return value
        raise ValueError(f"Unknown KS domain subtype: {label}")

@unique
class ERDomainSubtype(Enum):
    """
    An Enum representing the subtypes of KS domain supported by RAIChU
    """
    S = 1
    R = 2
    UNKNOWN = 3

    @staticmethod
    def from_string(label: str) -> "ERDomainSubtype":
        for value in ERDomainSubtype:
            if str(value.name) == label:
                return value
        raise ValueError(f"Unknown ER domain subtype: {label}")



from typing import List
import os

from pikachu.drawing.drawing import Drawer
from pikachu.reactions.functional_groups import find_atoms

from raichu.reactions.chain_release import cyclic_release, find_o_betapropriolactone
from raichu.drawing.drawer import RaichuDrawer
from raichu.module import _Module
from raichu.drawing.bubbles import draw_bubbles
from raichu.substrate import PKSSubstrate
from raichu.data.trans_at import (
    TRANSATOR_CLADE_TO_TAILORING_REACTIONS,
    TRANSATOR_CLADE_TO_STARTER_SUBSTRATE,
    TRANSATOR_CLADE_TO_ELONGATING,
)
from raichu.domain.domain import TailoringDomain
from raichu.data.molecular_moieties import O_OH, N_NH, KETO_GROUP_C
from raichu.cluster.base_cluster import Cluster
from raichu.representations import (
    TailoringRepresentation,
    MacrocyclizationRepresentation,
)


class ModularCluster(Cluster):
    def __init__(
        self,
        modules: List[_Module],
        tailoring_representations: List[TailoringRepresentation] = None,
    ) -> None:
        super().__init__(tailoring_representations)
        self.modules = modules
        self.tailoring_enzymes = tailoring_representations
        self.chain_intermediate = None

        self.tailored_product = None
        self.handle_transat()
        self.cyclisation_type = None
        self.modular = True

    def handle_transat(self):
        for i, module in enumerate(self.modules):
            if module.type.name == "PKS" and module.subtype.name == "PKS_TRANS":
                substrate = PKSSubstrate("MALONYL_COA")
                if module.is_starter_module:
                    substrate = PKSSubstrate("ACETYL_COA")
                if i < len(self.modules) - 1:
                    j = i + 1
                    next_module = self.modules[j]
                    if next_module.is_broken:
                        while next_module.is_broken and j < len(self.modules):
                            next_module = self.modules[j]
                            j += 1
                    if next_module.type.name == "PKS":
                        if (
                            module.is_starter_module
                            and next_module.subtype.name == "PKS_TRANS"
                        ):
                            if next_module.synthesis_domain:
                                if (
                                    next_module.synthesis_domain.subtype
                                    and next_module.synthesis_domain.subtype.name
                                    != "UNKNOWN"
                                ):
                                    substrate_name = (
                                        TRANSATOR_CLADE_TO_STARTER_SUBSTRATE.get(
                                            next_module.synthesis_domain.subtype.name
                                        )
                                    )
                                    if substrate_name is not None:
                                        substrate = PKSSubstrate(substrate_name)
                                    else:
                                        substrate = PKSSubstrate("ACETYL_COA")
                                else:
                                    substrate = PKSSubstrate("ACETYL_COA")
                            else:
                                substrate = PKSSubstrate("ACETYL_COA")
                                # TODO: Transfer names of substrates to a dictionary in raichu.substrate
                        elif module.is_starter_module:
                            substrate = PKSSubstrate("ACETYL_COA")
                        if next_module.synthesis_domain:
                            if (
                                not module.is_termination_module
                                and next_module.subtype.name == "PKS_TRANS"
                                and next_module.synthesis_domain.subtype
                                and next_module.synthesis_domain.subtype.name
                                != "UNKNOWN"
                                and next_module.synthesis_domain.subtype.name
                                != "MISCELLANEOUS"
                                and next_module.synthesis_domain.subtype.name
                                != "NON_ELONGATING"
                            ):
                                # Ignore tailoring domains in the module itself
                                module.tailoring_domains = []
                                for (
                                    dummy_domain_type,
                                    dummy_domain_subtype,
                                ) in TRANSATOR_CLADE_TO_TAILORING_REACTIONS[
                                    next_module.synthesis_domain.subtype.name
                                ]:
                                    self.modules[i].tailoring_domains.append(
                                        TailoringDomain(
                                            dummy_domain_type, dummy_domain_subtype
                                        )
                                    )
                if module.synthesis_domain:
                    if (
                        module.synthesis_domain.subtype
                        and module.synthesis_domain.subtype.name != "UNKNOWN"
                        and module.synthesis_domain.subtype.name != "MISCELLANEOUS"
                    ):
                        self.modules[i].synthesis_domain.is_elongating = (
                            TRANSATOR_CLADE_TO_ELONGATING[
                                module.synthesis_domain.subtype.name
                            ]
                        )
                self.modules[i].recognition_domain.substrate = substrate

    def compute_structures(self, compute_cyclic_products=True):

        for module in self.modules:
            structure = module.run_module(self.chain_intermediate)
            if module.carrier_domain and not module.is_broken:
                self.modular_intermediates.append(structure.deepcopy())
            self.chain_intermediate = structure
            if module.is_termination_module:
                if not structure:
                    raise ValueError("No product to release. Check if cluster has functional modules.")
                self.linear_product = module.release_chain(structure)
                self.chain_intermediate = self.linear_product
                if module.termination_domain.type.name in ["TE", "DUMMY_TE"]:
                    self.cyclisation_type = "condensative"
                elif module.termination_domain.type.name in ["TD", "DUMMY_TD"]:
                    self.cyclisation_type = "oxidative"
                else:
                    raise ValueError(
                        f"RAIChU does not recognise termination domain type {module.termination_domain.type.name}."
                    )
                break
        else:
            raise ValueError("Cluster must contain at least one termination module.")

        if compute_cyclic_products:
            self.cyclise_all()

    def cyclise(self, atom):

        cyclised_product = cyclic_release(self.chain_intermediate, atom)
        self.cyclic_product = cyclised_product
        return cyclised_product

    def cyclise_all(self):

        self.macrocyclisation_representations = []
        self.cyclised_products = []

        o_not_to_use = find_o_betapropriolactone(
            self.chain_intermediate, release_type=self.cyclisation_type
        )

        candidate_target_atoms = []
        terminal_o = None
        terminal_c = None

        for atom in find_atoms(N_NH, self.chain_intermediate):
            if len(atom.get_neighbours("H")) == 2:
                candidate_target_atoms.append(str(atom))

        if self.cyclisation_type == "condensative":
            for atom in find_atoms(O_OH, self.chain_intermediate):

                if (
                    atom.has_neighbour("H")
                    and atom != o_not_to_use
                    and not getattr(atom.annotations, "terminal_o")
                ):
                    candidate_target_atoms.append(str(atom))
                elif getattr(atom.annotations, "terminal_o"):
                    terminal_o = str(atom)

            assert terminal_o
            for atom in candidate_target_atoms:
                macrocyclisation_representation = MacrocyclizationRepresentation(
                    atom, terminal_o, "condensative"
                )
                self.macrocyclisation_representations.append(
                    macrocyclisation_representation
                )

        elif self.cyclisation_type == "oxidative":
            for atom in find_atoms(O_OH, self.chain_intermediate):
                if atom.has_neighbour("H") and atom != o_not_to_use:
                    candidate_target_atoms.append(str(atom))

            for atom in find_atoms(KETO_GROUP_C, self.chain_intermediate):
                if atom.has_neighbour("H") and getattr(atom.annotations, "terminal_c"):
                    terminal_c = str(atom)

            assert terminal_c

            for atom in candidate_target_atoms:
                macrocyclisation_representation = MacrocyclizationRepresentation(
                    atom, terminal_c, "oxidative"
                )
                self.macrocyclisation_representations.append(
                    macrocyclisation_representation
                )

        else:
            raise ValueError(
                f"Macrocyclisation must be oxidative or condensative. '{self.cyclisation_type}' given."
            )

        self.do_macrocyclization(sequential=False)

    def draw_all_products(self, out_dir):
        if not os.path.exists(out_dir):
            os.mkdir(out_dir)

        last_i = 0

        for i, product in enumerate(self.cyclic_products):
            out_file = os.path.join(out_dir, f"product_{i}.svg")
            drawing = Drawer(product)
            drawing.write_svg(out_file)
            last_i = i

        drawing = Drawer(self.linear_product)
        out_file = os.path.join(out_dir, f"product_{last_i + 1}.svg")
        drawing.write_svg(out_file)

    # TODO: Use PIKAChU's SVG drawer
    # def draw_spaghettis(self):
    #     spaghetti_svgs = []
    #     for structure in self.modular_intermediates:
    #         drawing = RaichuDrawer(structure, dont_show=True)
    #         drawing.draw_structure()
    #         svg_string = drawing.save_svg_string()
    #         spaghetti_svgs.append(svg_string)
    #
    #     linear_drawing = Drawer(self.linear_product)
    #     linear_svg = linear_drawing.save_svg_string()
    #
    #     return spaghetti_svgs + [linear_svg]

    def get_spaghettis(self, whitespace=30):
        drawings = []
        widths = []
        correct_modules = -1
        for i, module in enumerate(self.modules):

            if module.is_broken:
                widths.append((0, 0))
                continue

            correct_modules += 1

            structure = self.modular_intermediates[correct_modules]
            drawing = RaichuDrawer(structure, dont_show=True)

            drawing.flip_y_axis()
            drawing.finetune_overlap_bubbles()
            drawing.move_to_positive_coords()
            drawing.convert_to_int()

            carrier_domain_pos = None

            for atom in drawing.structure.graph:
                if atom.annotations.domain_type:
                    carrier_domain_pos = atom.draw.position
                    atom.draw.positioned = False

            assert carrier_domain_pos

            min_x = 100000000
            max_x = -100000000

            for atom in drawing.structure.graph:
                if atom.draw.positioned:
                    if atom.draw.position.x < min_x:
                        min_x = atom.draw.position.x
                    if atom.draw.position.x > max_x:
                        max_x = atom.draw.position.x

            width = (
                carrier_domain_pos.x - min_x + 0.5 * whitespace,
                max_x - carrier_domain_pos.x + 0.5 * whitespace,
            )
            widths.append(width)
            drawings.append(drawing)

        return drawings, widths

    def draw_cluster(self, as_string=True, out_file=None, colour_by_module=True):
        drawings, widths = self.get_spaghettis()
        bubble_svg, bubble_positions, last_domain_coord = draw_bubbles(self, widths)
        min_x = 100000000
        max_x = -100000000
        min_y = 100000000
        max_y = -100000000

        svg_strings = []
        squiggly_svgs = []
        padding = None
        svg_style = None

        for i, drawing in enumerate(drawings):
            drawing.set_structure_id(f"s{i}")
            drawing.set_annotation_for_grouping("module_nr")
            if colour_by_module:
                drawing.colour_by_annotation()
            svg_style = drawing.svg_style

            padding = drawing.options.padding

            carrier_domain_pos = None

            for atom in drawing.structure.graph:
                if atom.annotations.domain_type:
                    carrier_domain_pos = atom.draw.position
                    atom.draw.positioned = False

            assert carrier_domain_pos
            bubble_x, bubble_y = bubble_positions[i]
            x_translation = bubble_x - carrier_domain_pos.x
            y_translation = bubble_y - carrier_domain_pos.y

            drawing.move_structure(x_translation, y_translation + 15)
            svg = drawing.draw_svg(annotation="module_nr")
            svg_strings.append(svg)

            sulphur_pos = None
            carrier_pos = None

            for atom in drawing.structure.graph:
                if atom.draw.positioned:
                    if atom.draw.position.x < min_x:
                        min_x = atom.draw.position.x
                    if atom.draw.position.y < min_y:
                        min_y = atom.draw.position.y
                    if atom.draw.position.x > max_x:
                        max_x = atom.draw.position.x
                    if atom.draw.position.y > max_y:
                        max_y = atom.draw.position.y
                if atom.annotations.domain_type:
                    sulphur_pos = atom.get_neighbour("S").draw.position
                    carrier_pos = atom.draw.position

            squiggly_svg = f'<path d="M {sulphur_pos.x} {sulphur_pos.y - 5} Q {sulphur_pos.x - 5} {sulphur_pos.y - (sulphur_pos.y - 5 - carrier_pos.y)/2}, {carrier_pos.x} {sulphur_pos.y - 5 - (sulphur_pos.y - 5 - carrier_pos.y)/2} T {carrier_pos.x} {carrier_pos.y}" stroke="grey" fill="white"/>'
            squiggly_svgs.append(squiggly_svg)

        assert padding is not None

        x1 = 0
        x2 = max([max_x + padding, last_domain_coord + padding])
        y1 = 0
        y2 = max_y + padding

        width = x2
        height = y2

        svg_string = f"""<svg width="{width}" height="{height}" viewBox="{x1} {y1} {x2} {y2}" xmlns="http://www.w3.org/2000/svg">\n"""
        if svg_style:
            svg_string += f"{svg_style}\n"
        svg_string += bubble_svg
        for string in svg_strings:
            svg_string += string
        for squiggly_svg in squiggly_svgs:
            svg_string += squiggly_svg
        svg_string += "</svg>"

        if as_string:
            return svg_string
        else:
            if out_file is None:
                raise ValueError(
                    "Must provide output svg directory if 'as_string' is set to False."
                )
            else:
                with open(out_file, "w") as svg_out:
                    svg_out.write(svg_string)



import pygame

from interactive.style import BUTTON_PANEL_COLOUR, BUTTON_PANEL, BLACK, SIZE, RENDER_WINDOW_SIZE


class ButtonPanel:
    def __init__(self, screen):
        self.screen = screen

    def draw(self):
        # pygame.draw.rect(self.screen, BUTTON_PANEL_COLOUR, BUTTON_PANEL)
        pygame.draw.line(self.screen, BLACK, (BUTTON_PANEL.x, BUTTON_PANEL.y),
                         (BUTTON_PANEL.x + BUTTON_PANEL.width, BUTTON_PANEL.y))


class ProductRaster:
    def __init__(self, screen):
        self.screen = screen
        self.rectangles = self.get_rectangles()

    def get_rectangles(self):
        panel_width = RENDER_WINDOW_SIZE[0] / 3
        panel_height = RENDER_WINDOW_SIZE[1] / 2
        x_1 = SIZE[0] - RENDER_WINDOW_SIZE[0]
        x_2 = x_1 + panel_width
        x_3 = x_2 + panel_width
        y_1 = 0
        y_2 = panel_height

        rectangles = [pygame.Rect(x_1, y_1, panel_width, panel_height),
                      pygame.Rect(x_2, y_1, panel_width, panel_height),
                      pygame.Rect(x_3, y_1, panel_width, panel_height),
                      pygame.Rect(x_1, y_2, panel_width, panel_height),
                      pygame.Rect(x_2, y_2, panel_width, panel_height),
                      pygame.Rect(x_3, y_2, panel_width, panel_height)]

        return rectangles

    def draw_images(self, images):
        for i, image in enumerate(images):
            if i <= 5:
                center = self.rectangles[i].center
                rectangle = image.get_rect(center=center)
                self.screen.blit(image, rectangle)
                pygame.draw.rect(self.screen, BLACK, rectangle, 2)







import pygame
import os

from interactive.style import BUTTON_TEXT_COLOUR, FONT, BUTTON_HIGHLIGHT_COLOUR, BLACK, BUTTON_COLOUR, \
    BUTTON_PANEL_COLOUR, HEIGHT, WIDTH, DOMAIN_BUTTON_SIZE, SUBSTRATE_GROUP_BUTTONS_PER_LINE, \
    SUBSTRATE_GROUP_BUTTON_PADDING, SUBSTRATE_GROUP_BUTTON_SIZE, SUBSTRATE_BUTTON_SIZE, KR_BUTTON_SIZE
import interactive.images.domains
import interactive.flatfiles
import interactive.images.kr_subtypes
from interactive.gene import Gene
from interactive.textbox import TextBox
from interactive.parsers import parse_smiles
from interactive.substrate import Substrate, SubstrateGroup, \
    PROTEINOGENIC_SUBSTRATES, NONPROTEINOGENIC_SUBSTRATES, FATTY_ACIDS, NON_AMINO_ACIDS

DOMAIN_IMAGE_DIR = os.path.dirname(interactive.images.domains.__file__)
KR_IMAGE_DIR = os.path.dirname(interactive.images.kr_subtypes.__file__)
FLATFILES = os.path.dirname(interactive.flatfiles.__file__)
PARAS_SMILES = os.path.join(FLATFILES, "PARAS_smiles.txt")
PKS_SMILES = os.path.join(FLATFILES, "at_specificities.txt")


class Button:
    def __init__(self, text, position, dimensions, font_size=None):

        self.text = text

        self.x, self.y = position
        self.width, self.height = dimensions
        self.rectangle = pygame.Rect(position, dimensions)
        self.font_size = int(self.height) - 10

        self.text_x = 0
        self.text_y = 0

        self.set_text_position()
        self.font = pygame.font.SysFont(FONT, self.font_size, bold=False)

        self.rendered_text = self.font.render(text, True, BUTTON_TEXT_COLOUR)
        self.selected = False

    def __hash__(self):
        return hash(self.text)

    def __eq__(self, other):
        return self.text == other.text

    def __repr__(self):
        return self.text

    def highlight(self, screen):
        pygame.draw.rect(screen, BUTTON_HIGHLIGHT_COLOUR, self.rectangle)
        pygame.draw.rect(screen, BLACK, self.rectangle, 2)
        screen.blit(self.rendered_text, (self.text_x, self.text_y))

    def draw(self, screen):
        pygame.draw.rect(screen, BUTTON_COLOUR, self.rectangle)
        pygame.draw.rect(screen, BLACK, self.rectangle, 2)
        screen.blit(self.rendered_text, (self.text_x, self.text_y))

    def erase_button(self, screen):
        pygame.draw.rect(screen, BUTTON_PANEL_COLOUR, self.rectangle)
        pygame.draw.rect(screen, BUTTON_PANEL_COLOUR, self.rectangle, 2)

    def set_text_position(self):
        self.text_x = self.x + 5
        self.text_y = self.y + int((self.height - self.font_size) / 2)

    def set_font(self):
        self.font = pygame.font.SysFont(FONT, self.font_size, bold=True)


class RenderClusterButton(Button):

    def __init__(self):
        position = (int(0.77 * WIDTH), int(0.92 * HEIGHT))
        dimensions = (int(0.2 * WIDTH), int(HEIGHT / 25))

        super().__init__("Render cluster", position, dimensions)

    def do_action(self, screen, active_buttons):
        position = (int(0.77 * WIDTH), int(0.92 * HEIGHT))
        dimensions = (int(0.2 * WIDTH), int(HEIGHT / 25))
        text_box = TextBox(position, dimensions)
        text_box.draw(screen)
        hide_buttons(ALL_BUTTONS, screen, active_buttons)
        show_button(SAVE_CLUSTER_BUTTON, screen, active_buttons)
        return text_box


class KRSubtypeButton(Button):
    def __init__(self, position, kr_subtype):
        self.kr_subtype = kr_subtype
        dimensions = (KR_BUTTON_SIZE, KR_BUTTON_SIZE)

        super().__init__(kr_subtype, position, dimensions)

        if kr_subtype:

            self.image = os.path.join(KR_IMAGE_DIR, f"KR_{kr_subtype}.png")
            self.highlight_image = os.path.join(KR_IMAGE_DIR, f"KR_{kr_subtype}_highlight.png")
        else:
            self.image = os.path.join(KR_IMAGE_DIR, f"KR.png")
            self.highlight_image = os.path.join(KR_IMAGE_DIR, f"KR_highlight.png")


    def draw(self, screen):
        kr_image = pygame.image.load(self.image)
        kr_image_scaled = pygame.transform.smoothscale(kr_image, (KR_BUTTON_SIZE, KR_BUTTON_SIZE))
        screen.blit(kr_image_scaled, self.rectangle)

    def highlight(self, screen):
        kr_image = pygame.image.load(self.highlight_image)
        kr_image_scaled = pygame.transform.smoothscale(kr_image, (KR_BUTTON_SIZE, KR_BUTTON_SIZE))
        screen.blit(kr_image_scaled, self.rectangle)

    def do_action(self, domain, mouse, screen, active_buttons):
        if self.kr_subtype:
            domain.subtype = self.kr_subtype
        else:
            domain.subtype = None

        domain.module.gene.erase()
        domain.module.gene.draw(mouse)
        reset_buttons(screen, active_buttons)


class DomainButton(Button):
    def __init__(self, position, text):
        dimensions = (DOMAIN_BUTTON_SIZE, DOMAIN_BUTTON_SIZE)
        super().__init__(text, position, dimensions)
        self.image = None
        self.highlight_image = None
        self.domain_type = text

        self.image = os.path.join(DOMAIN_IMAGE_DIR, f'{text.lower()}.png')
        self.highlight_image = os.path.join(DOMAIN_IMAGE_DIR, f'{text.lower()}_highlight.png')

    def draw(self, screen):
        domain_image = pygame.image.load(self.image)
        domain_image_scaled = pygame.transform.smoothscale(domain_image, (DOMAIN_BUTTON_SIZE, DOMAIN_BUTTON_SIZE))
        screen.blit(domain_image_scaled, self.rectangle)

    def highlight(self, screen):
        domain_image = pygame.image.load(self.highlight_image)
        domain_image_scaled = pygame.transform.smoothscale(domain_image, (DOMAIN_BUTTON_SIZE, DOMAIN_BUTTON_SIZE))
        screen.blit(domain_image_scaled, self.rectangle)

    def do_action(self, module, mouse, screen, active_buttons):
        module.gene.erase()
        module.add_domain(self.domain_type)
        module.gene.draw(mouse)
        reset_buttons(screen, active_buttons)


class AddGeneButton(Button):
    def __init__(self):
        position = (int(0.02 * WIDTH), int(0.57 * HEIGHT))
        dimensions = (int(0.2 * WIDTH), int(HEIGHT / 25))

        super().__init__("Add gene", position, dimensions)

    def do_action(self, screen, genes, text_box, active_buttons, mouse):
        gene_number = len(genes)
        gene = Gene(screen, gene_number)
        gene.name = text_box.text
        text_box.erase(screen)
        gene.draw(mouse)
        genes.append(gene)
        reset_buttons(screen, active_buttons)


class RemoveModuleButton(Button):

    def __init__(self):
        position = (int(0.02 * WIDTH), int(0.57 * HEIGHT))
        dimensions = (int(0.2 * WIDTH), int(HEIGHT / 25))

        super().__init__("Remove module", position, dimensions)

    def do_action(self, screen, current_module, active_buttons, mouse):
        gene = current_module.gene
        gene.remove_module(current_module.id)

        gene.erase()
        gene.draw(mouse)

        reset_buttons(screen, active_buttons)


class RemoveGeneButton(Button):
    def __init__(self):
        position = (int(0.02 * WIDTH), int(0.57 * HEIGHT))
        dimensions = (int(0.2 * WIDTH), int(HEIGHT / 25))

        super().__init__("Remove gene", position, dimensions)

    def do_action(self, screen, genes, current_gene, active_buttons, mouse):
        gene_number = current_gene.gene_number
        index_to_remove = None
        for i, gene in enumerate(genes):
            if gene.gene_number == gene_number:
                index_to_remove = i
            elif gene.gene_number > gene_number:
                gene.gene_number -= 1

        genes.pop(index_to_remove)
        current_gene.erase()

        for gene in genes:
            gene.erase()
            gene.draw(mouse)

        reset_buttons(screen, active_buttons)


class CreateGeneButton(Button):
    def __init__(self):
        position = (int(0.02 * WIDTH), int(0.52 * HEIGHT))
        dimensions = (int(0.2 * WIDTH), int(HEIGHT / 25))

        super().__init__("Create gene", position, dimensions)

    def do_action(self, screen, active_buttons):
        position = (int(0.02 * WIDTH), int(0.52 * HEIGHT))
        dimensions = (int(0.2 * WIDTH), int(HEIGHT / 25))
        text_box = TextBox(position, dimensions)
        text_box.draw(screen)
        hide_buttons(ALL_BUTTONS, screen, active_buttons)
        show_button(ADD_GENE_BUTTON, screen, active_buttons)
        return text_box


class SubstrateGroupButton(Button):
    def __init__(self, text, position, dimensions):
        super().__init__(text, position, dimensions)


class AddModuleButton(Button):
    def __init__(self):
        position = (int(0.02 * WIDTH), int(0.62 * HEIGHT))
        dimensions = (int(0.2 * WIDTH), int(HEIGHT / 25))

        super().__init__("Add module", position, dimensions)


class NRPSModuleButton(Button):
    def __init__(self):
        position = (int(0.02 * WIDTH), int(0.67 * HEIGHT))
        dimensions = (int(0.09 * WIDTH), int(HEIGHT / 25))

        super().__init__('NRPS', position, dimensions)

    def do_action(self, screen, current_gene, active_buttons, insertion_point, mouse):
        module_nr = insertion_point.insertion_point

        current_gene.add_module(module_nr, 'NRPS')
        current_gene.erase()
        current_gene.insertion_point = None
        current_gene.draw(mouse)
        reset_buttons(screen, active_buttons)


class PKSModuleButton(Button):
    def __init__(self):
        position = (int(0.13 * WIDTH), int(0.67 * HEIGHT))
        dimensions = (int(0.09 * WIDTH), int(HEIGHT / 25))

        super().__init__('PKS', position, dimensions)

    def do_action(self, screen, current_gene, active_buttons, insertion_point, mouse):
        module_nr = insertion_point.insertion_point

        current_gene.add_module(module_nr, 'PKS')
        current_gene.erase()
        current_gene.insertion_point = None
        current_gene.draw(mouse)
        reset_buttons(screen, active_buttons)


class AddDomainButton(Button):
    def __init__(self):
        position = (int(0.02 * WIDTH), int(0.62 * HEIGHT))
        dimensions = (int(0.2 * WIDTH), int(HEIGHT / 25))

        super().__init__("Add domain", position, dimensions)


class ACPDomainButton(DomainButton):
    def __init__(self):
        position = (int(0.35 * WIDTH), int(0.52 * HEIGHT))

        super().__init__(position, 'ACP')


class PCPDomainButton(DomainButton):
    def __init__(self):
        position = (int(0.35 * WIDTH), int(0.52 * HEIGHT))
        super().__init__(position, 'PCP')


class ADomainButton(DomainButton):
    def __init__(self):
        position = (int(0.30 * WIDTH), int(0.52 * HEIGHT))
        super().__init__(position, 'A')


class CDomainButton(DomainButton):
    def __init__(self):
        position = (int(0.25 * WIDTH), int(0.52 * HEIGHT))
        super().__init__(position, 'C')


class KSDomainButton(DomainButton):
    def __init__(self):
        position = (int(0.25 * WIDTH), int(0.52 * HEIGHT))
        super().__init__(position, 'KS')


class KRDomainButton(DomainButton):
    def __init__(self):
        position = (int(0.25 * WIDTH), int(0.62 * HEIGHT))
        super().__init__(position, 'KR')


class DHDomainButton(DomainButton):
    def __init__(self):
        position = (int(0.30 * WIDTH), int(0.62 * HEIGHT))
        super().__init__(position, 'DH')


class ERDomainButton(DomainButton):
    def __init__(self):
        position = (int(0.35 * WIDTH), int(0.62 * HEIGHT))
        super().__init__(position, 'ER')


class EDomainButton(DomainButton):
    def __init__(self):
        position = (int(0.25 * WIDTH), int(0.62 * HEIGHT))
        super().__init__(position, 'E')


class NMTDomainButton(DomainButton):
    def __init__(self):
        position = (int(0.30 * WIDTH), int(0.62 * HEIGHT))
        super().__init__(position, 'nMT')


class ATDomainButton(DomainButton):
    def __init__(self):
        position = (int(0.30 * WIDTH), int(0.52 * HEIGHT))
        super().__init__(position, 'AT')


class TEDomainButton(DomainButton):
    def __init__(self):
        position = (int(0.25 * WIDTH), int(0.72 * HEIGHT))
        super().__init__(position, 'TE')


class RemoveDomainButton(Button):
    def __init__(self):
        position = (int(0.02 * WIDTH), int(0.57 * HEIGHT))
        dimensions = (int(0.2 * WIDTH), int(HEIGHT / 25))

        super().__init__("Remove domain", position, dimensions)

    def do_action(self, current_domain, screen, active_buttons, mouse):
        gene = current_domain.module.gene
        current_domain.module.remove_domain(current_domain)
        gene.erase()
        gene.draw(mouse)
        reset_buttons(screen, active_buttons)


class SelectSubstrateButton(Button):
    def __init__(self):
        position = (int(0.75 * WIDTH), int(0.52 * HEIGHT))
        dimensions = (int(0.2 * WIDTH), int(HEIGHT / 25))

        super().__init__("Select substrate", position, dimensions)


class SelectDomainTypeButton(Button):
    def __init__(self):
        position = (int(0.75 * WIDTH), int(0.57 * HEIGHT))
        dimensions = (int(0.2 * WIDTH), int(HEIGHT / 25))

        super().__init__("Select subtype", position, dimensions)


class SetDomainInactiveButton(Button):

    def __init__(self):
        position = (int(0.75 * WIDTH), int(0.62 * HEIGHT))
        dimensions = (int(0.2 * WIDTH), int(HEIGHT / 25))

        super().__init__("Set to inactive", position, dimensions)


class SubstrateButton(Button):
    def __init__(self, substrate, position, dimensions):

        super().__init__(substrate.name, position, dimensions)
        self.substrate = substrate
        self.image = substrate.image
        self.highlight_image = substrate.highlight_image

    def draw(self, screen):
        substrate_image = pygame.image.load(self.image)
        substrate_image_scaled = pygame.transform.smoothscale(substrate_image, (SUBSTRATE_BUTTON_SIZE,
                                                                                SUBSTRATE_BUTTON_SIZE))
        screen.blit(substrate_image_scaled, self.rectangle)

    def highlight(self, screen):
        substrate_image = pygame.image.load(self.highlight_image)
        substrate_image_scaled = pygame.transform.smoothscale(substrate_image, (SUBSTRATE_BUTTON_SIZE,
                                                                                SUBSTRATE_BUTTON_SIZE))
        screen.blit(substrate_image_scaled, self.rectangle)

    def do_action(self):
        pass


class NRPSWildcardButton(SubstrateButton):
    def __init__(self):
        substrate = Substrate('nrp', 'NC(C(=O)O)[*]', 'NRPS')
        position = (int(0.52 * WIDTH), int(0.52 * HEIGHT))
        substrate.x = position[0]
        substrate.y = position[1]
        substrate.rectangle = pygame.Rect(substrate.x, substrate.y, substrate.width, substrate.height)
        dimensions = (substrate.width, substrate.height)

        super().__init__(substrate, position, dimensions)


class SubstrateSupergroupButton(Button):
    def __init__(self, text, position, group):
        dimensions = (int(0.2 * WIDTH), int(HEIGHT / 25))
        self.group = group

        super().__init__(text, position, dimensions)


class ProteinogenicButton(SubstrateSupergroupButton):
    def __init__(self):
        position = (int(0.25 * WIDTH), int(0.52 * HEIGHT))

        super().__init__("Proteinogenic", position,
                         PROTEINOGENIC_SUBSTRATES)


class NonProteinogenicButton(SubstrateSupergroupButton):
    def __init__(self):
        position = (int(0.25 * WIDTH), int(0.57 * HEIGHT))

        super().__init__("Non-proteinogenic", position,
                         NONPROTEINOGENIC_SUBSTRATES)


class NonAminoAcidButton(SubstrateSupergroupButton):
    def __init__(self):
        position = (int(0.25 * WIDTH), int(0.52 * HEIGHT))

        super().__init__("Non-amino acids", position,
                         NON_AMINO_ACIDS)


class FattyAcidButton(SubstrateSupergroupButton):
    def __init__(self):
        position = (int(0.25 * WIDTH), int(0.57 * HEIGHT))

        super().__init__("Fatty acids", position,
                         FATTY_ACIDS)


class RenderProductsButton(Button):
    def __init__(self):
        position = (int(0.52 * WIDTH), int(0.92 * HEIGHT))
        dimensions = (int(0.2 * WIDTH), int(HEIGHT / 25))

        super().__init__("Render products", position, dimensions)

    def do_action(self, screen, active_buttons):
        position = (int(0.02 * WIDTH), int(0.92 * HEIGHT))
        dimensions = (int(0.2 * WIDTH), int(HEIGHT / 25))
        text_box = TextBox(position, dimensions)
        text_box.draw(screen)
        hide_buttons(ALL_BUTTONS, screen, active_buttons)
        show_button(SAVE_PRODUCTS_BUTTON, screen, active_buttons)
        return text_box


class YesButton(Button):
    def __init__(self):
        position = (int(0.02 * WIDTH), int(0.92 * HEIGHT))
        dimensions = (int(0.05 * WIDTH), int(HEIGHT / 25))

        super().__init__("Yes", position, dimensions)


class NoButton(Button):
    def __init__(self):
        position = (int(0.08 * WIDTH), int(0.92 * HEIGHT))
        dimensions = (int(0.05 * WIDTH), int(HEIGHT / 25))

        super().__init__("No", position, dimensions)


class FattyAcidSuperOptionButton(Button):
    def __init__(self, text, position, dimensions):
        super().__init__(text, position, dimensions)


class FattyAcidOptionButton(Button):
    def __init__(self, text, position, dimensions):
        super().__init__(text, position, dimensions)


class SaveClusterButton(Button):
    def __init__(self):
        position = (int(0.02 * WIDTH), int(0.82 * HEIGHT))
        dimensions = (int(0.2 * WIDTH), int(HEIGHT / 25))

        super().__init__("Save cluster", position, dimensions)

    def do_action(self, screen, active_buttons):
        position = (int(0.02 * WIDTH), int(0.87 * HEIGHT))
        dimensions = (int(0.2 * WIDTH), int(HEIGHT / 25))
        text_box = TextBox(position, dimensions)
        text_box.draw(screen)
        hide_buttons(ALL_BUTTONS, screen, active_buttons)
        show_button(SAVE_TO_PNG_BUTTON, screen, active_buttons)
        return text_box


class ExportTabularButton(Button):
    def __init__(self):
        position = (int(0.27 * WIDTH), int(0.92 * HEIGHT))
        dimensions = (int(0.2 * WIDTH), int(HEIGHT / 25))

        super().__init__("Export to tabular", position, dimensions)

    def do_action(self, screen, active_buttons):
        position = (int(0.02 * WIDTH), int(0.87 * HEIGHT))
        dimensions = (int(0.2 * WIDTH), int(HEIGHT / 25))
        text_box = TextBox(position, dimensions)
        text_box.draw(screen)
        hide_buttons(ALL_BUTTONS, screen, active_buttons)
        show_button(SAVE_TO_TXT_BUTTON, screen, active_buttons)
        return text_box


class SaveToTxtButton(Button):
    def __init__(self):
        position = (int(0.02 * WIDTH), int(0.82 * HEIGHT))
        dimensions = (int(0.2 * WIDTH), int(HEIGHT / 25))

        super().__init__("Save to tabular", position, dimensions)


class SaveToPngButton(Button):
    def __init__(self):
        position = (int(0.02 * WIDTH), int(0.82 * HEIGHT))
        dimensions = (int(0.2 * WIDTH), int(HEIGHT / 25))

        super().__init__("Save to png", position, dimensions)


class CButton(Button):
    def __init__(self, nr, position, dimensions):
        self.nr = nr
        super().__init__(f"C{nr}", position, dimensions)


class SaveToFolderButton(Button):
    def __init__(self):
        position = (int(0.02 * WIDTH), int(0.82 * HEIGHT))
        dimensions = (int(0.2 * WIDTH), int(HEIGHT / 25))

        super().__init__("Save to folder", position, dimensions)


class SaveProductsButton(Button):
    def __init__(self):
        position = (int(0.02 * WIDTH), int(0.82 * HEIGHT))
        dimensions = (int(0.2 * WIDTH), int(HEIGHT / 25))

        super().__init__("Save products", position, dimensions)

    def do_action(self, screen, active_buttons):
        position = (int(0.02 * WIDTH), int(0.87 * HEIGHT))
        dimensions = (int(0.2 * WIDTH), int(HEIGHT / 25))
        text_box = TextBox(position, dimensions)
        text_box.draw(screen)
        hide_buttons(ALL_BUTTONS, screen, active_buttons)
        show_button(SAVE_TO_FOLDER_BUTTON, screen, active_buttons)
        return text_box


class StarterButton(Button):
    def __init__(self):
        position = (int(0.25 * WIDTH), int(0.52 * HEIGHT))
        dimensions = (int(0.2 * WIDTH), int(HEIGHT / 25))

        super().__init__("Starter substrates", position,
                         dimensions)


class ElongationButton(Button):
    def __init__(self):
        position = (int(0.25 * WIDTH), int(0.57 * HEIGHT))
        dimensions = (int(0.2 * WIDTH), int(HEIGHT / 25))

        super().__init__("Elongation substrates", position,
                         dimensions)


RENDER_CLUSTER_BUTTON = RenderClusterButton()
RENDER_PRODUCTS_BUTTON = RenderProductsButton()
SAVE_CLUSTER_BUTTON = SaveClusterButton()
SAVE_PRODUCTS_BUTTON = SaveProductsButton()
SAVE_TO_FOLDER_BUTTON = SaveToFolderButton()
SAVE_TO_PNG_BUTTON = SaveToPngButton()
SAVE_TO_TXT_BUTTON = SaveToTxtButton()
EXPORT_TABULAR_BUTTON = ExportTabularButton()

YES_BUTTON = YesButton()
NO_BUTTON = NoButton()

CREATE_GENE_BUTTON = CreateGeneButton()
ADD_GENE_BUTTON = AddGeneButton()
REMOVE_GENE_BUTTON = RemoveGeneButton()

ADD_MODULE_BUTTON = AddModuleButton()
REMOVE_MODULE_BUTTON = RemoveModuleButton()

NRPS_MODULE_BUTTON = NRPSModuleButton()
PKS_MODULE_BUTTON = PKSModuleButton()

ADD_DOMAIN_BUTTON = AddDomainButton()
REMOVE_DOMAIN_BUTTON = RemoveDomainButton()

SELECT_SUBSTRATE_BUTTON = SelectSubstrateButton()
SELECT_DOMAIN_TYPE_BUTTON = SelectDomainTypeButton()
SET_DOMAIN_INACTIVE_BUTTON = SetDomainInactiveButton()


KS_DOMAIN_BUTTON = KSDomainButton()
AT_DOMAIN_BUTTON = ATDomainButton()
ACP_DOMAIN_BUTTON = ACPDomainButton()
DH_DOMAIN_BUTTON = DHDomainButton()
ER_DOMAIN_BUTTON = ERDomainButton()
KR_DOMAIN_BUTTON = KRDomainButton()

C_DOMAIN_BUTTON = CDomainButton()
A_DOMAIN_BUTTON = ADomainButton()
PCP_DOMAIN_BUTTON = PCPDomainButton()
E_DOMAIN_BUTTON = EDomainButton()
NMT_DOMAIN_BUTTON = NMTDomainButton()
TE_DOMAIN_BUTTON = TEDomainButton()

STARTER_BUTTON = StarterButton()
ELONGATION_BUTTON = ElongationButton()

PKS_DOMAIN_TO_BUTTON = {'KS': KS_DOMAIN_BUTTON,
                        'AT': AT_DOMAIN_BUTTON,
                        'DH': DH_DOMAIN_BUTTON,
                        'ER': ER_DOMAIN_BUTTON,
                        'KR': KR_DOMAIN_BUTTON,
                        'ACP': ACP_DOMAIN_BUTTON,
                        'TE': TE_DOMAIN_BUTTON}

NRPS_DOMAIN_TO_BUTTON = {'C': C_DOMAIN_BUTTON,
                         'A': A_DOMAIN_BUTTON,
                         'nMT': NMT_DOMAIN_BUTTON,
                         'PCP': PCP_DOMAIN_BUTTON,
                         'E': E_DOMAIN_BUTTON,
                         'TE': TE_DOMAIN_BUTTON}

PROTEINOGENIC_BUTTON = ProteinogenicButton()
NON_PROTEINOGENIC_BUTTON = NonProteinogenicButton()
NON_AMINO_ACID_BUTTON = NonAminoAcidButton()
FATTY_ACID_BUTTON = FattyAcidButton()
NRPS_WILDCARD_BUTTON = NRPSWildcardButton()

NRPS_SUPERGROUP_BUTTONS = [PROTEINOGENIC_BUTTON,
                           NON_PROTEINOGENIC_BUTTON,
                           NON_AMINO_ACID_BUTTON,
                           FATTY_ACID_BUTTON,
                           NRPS_WILDCARD_BUTTON]

ALL_BUTTONS = [RENDER_CLUSTER_BUTTON,
               RENDER_PRODUCTS_BUTTON,
               CREATE_GENE_BUTTON,
               ADD_GENE_BUTTON,
               ADD_MODULE_BUTTON,
               NRPS_MODULE_BUTTON,
               PKS_MODULE_BUTTON,
               KS_DOMAIN_BUTTON,
               AT_DOMAIN_BUTTON,
               ACP_DOMAIN_BUTTON,
               DH_DOMAIN_BUTTON,
               ER_DOMAIN_BUTTON,
               KR_DOMAIN_BUTTON,
               C_DOMAIN_BUTTON,
               A_DOMAIN_BUTTON,
               PCP_DOMAIN_BUTTON,
               E_DOMAIN_BUTTON,
               TE_DOMAIN_BUTTON,
               NMT_DOMAIN_BUTTON,
               ADD_DOMAIN_BUTTON,
               REMOVE_DOMAIN_BUTTON,
               REMOVE_MODULE_BUTTON,
               REMOVE_GENE_BUTTON,
               SELECT_SUBSTRATE_BUTTON,
               SELECT_DOMAIN_TYPE_BUTTON,
               SET_DOMAIN_INACTIVE_BUTTON,
               PROTEINOGENIC_BUTTON,
               NON_PROTEINOGENIC_BUTTON,
               NON_AMINO_ACID_BUTTON,
               FATTY_ACID_BUTTON,
               SAVE_CLUSTER_BUTTON,
               SAVE_PRODUCTS_BUTTON,
               NRPS_WILDCARD_BUTTON,
               SAVE_TO_FOLDER_BUTTON,
               SAVE_TO_PNG_BUTTON,
               YES_BUTTON,
               NO_BUTTON,
               STARTER_BUTTON,
               ELONGATION_BUTTON,
               EXPORT_TABULAR_BUTTON,
               SAVE_TO_TXT_BUTTON]


def show_domain_buttons(module, screen, active_buttons):
    reset_buttons(screen, active_buttons)
    domain_types = set()

    for domain in module.domains:
        domain_types.add(domain.type)

    if module.type == 'NRPS':

        for domain, button in NRPS_DOMAIN_TO_BUTTON.items():
            if domain not in domain_types:
                show_button(button, screen, active_buttons)

    elif module.type == 'PKS':

        for domain, button in PKS_DOMAIN_TO_BUTTON.items():
            if domain not in domain_types:
                show_button(button, screen, active_buttons)


def hide_domain_buttons(screen, active_buttons):
    for button in active_buttons:
        if type(button) == DomainButton:
            hide_button(button, screen, active_buttons)


def highlight_buttons(active_buttons, screen, mouse):
    for button in active_buttons:
        if button.rectangle.collidepoint(mouse) or button.selected:
            button.highlight(screen)
        else:
            button.draw(screen)


def get_mouse_button(active_buttons, mouse):
    for button in active_buttons:
        if button.rectangle.collidepoint(mouse):
            return button

    return None


def make_buttons(screen):
    buttons = set()
    buttons.add(CREATE_GENE_BUTTON)
    CREATE_GENE_BUTTON.draw(screen)

    buttons.add(RENDER_CLUSTER_BUTTON)
    RENDER_CLUSTER_BUTTON.draw(screen)

    buttons.add(RENDER_PRODUCTS_BUTTON)
    RENDER_PRODUCTS_BUTTON.draw(screen)

    buttons.add(EXPORT_TABULAR_BUTTON)
    EXPORT_TABULAR_BUTTON.draw(screen)

    return buttons


def draw_buttons(active_buttons, screen, mouse):
    for button in active_buttons:
        if button.rectangle.collidepoint(mouse) or button.selected:
            button.highlight(screen)
        else:
            button.draw(screen)


def reset_buttons(screen, active_buttons):
    hide_buttons(ALL_BUTTONS, screen, active_buttons)
    show_button(CREATE_GENE_BUTTON, screen, active_buttons)
    show_button(RENDER_CLUSTER_BUTTON, screen, active_buttons)
    show_button(RENDER_PRODUCTS_BUTTON, screen, active_buttons)
    show_button(EXPORT_TABULAR_BUTTON, screen, active_buttons)


def hide_button(button, screen, active_buttons):
    try:
        active_buttons.remove(button)
    except KeyError:
        pass

    button.erase_button(screen)


def hide_buttons(buttons, screen, active_buttons):
    for button in buttons:
        hide_button(button, screen, active_buttons)


def show_button(button, screen, active_buttons):
    button.draw(screen)
    active_buttons.add(button)


def show_buttons(buttons, screen, active_buttons):
    for button in buttons:
        show_button(button, screen, active_buttons)


def make_raster(substrate_groups, buttons, buttons_per_line, button_size):
    x_coord = int(0.25 * WIDTH)
    y_coord = int(0.52 * HEIGHT)

    for i, group in enumerate(substrate_groups):
        if i % buttons_per_line == 0:
            if i != 0:
                y_coord += int(0.05 * HEIGHT)
            x_coord = int(0.25 * WIDTH)
        else:
            x_coord += button_size[0] + SUBSTRATE_GROUP_BUTTON_PADDING

        button = SubstrateGroupButton(group, (x_coord, y_coord), button_size)
        buttons.add(button)


def make_nrps_substrate_group_buttons():
    proteinogenic_buttons = set()
    nonproteinogenic_buttons = set()
    fatty_acid_buttons = set()
    non_amino_acid_buttons = set()

    make_raster(PROTEINOGENIC_SUBSTRATES, proteinogenic_buttons, SUBSTRATE_GROUP_BUTTONS_PER_LINE, SUBSTRATE_GROUP_BUTTON_SIZE)
    make_raster(NONPROTEINOGENIC_SUBSTRATES, nonproteinogenic_buttons, SUBSTRATE_GROUP_BUTTONS_PER_LINE, (int(WIDTH * 0.08), int(HEIGHT / 25)))
    make_raster(FATTY_ACIDS, fatty_acid_buttons, 1, (int(WIDTH * 0.15), int(HEIGHT / 25)))
    make_raster(NON_AMINO_ACIDS, non_amino_acid_buttons, 1, (int(WIDTH * 0.15), int(HEIGHT / 25)))

    return [proteinogenic_buttons, nonproteinogenic_buttons, fatty_acid_buttons, non_amino_acid_buttons]


def make_nrps_substrate_groups(supergroup, name_to_smiles, starter=False):
    group_to_buttons = {}
    for group, substrate_names in supergroup.items():
        group_to_buttons[group] = set()

        substrates = []

        for substrate_name in substrate_names:
            smiles = name_to_smiles[substrate_name]
            substrate = Substrate(substrate_name, smiles, 'NRPS', starter=starter)
            substrates.append(substrate)

        substrate_group = SubstrateGroup(group, substrates)
        for substrate in substrate_group.substrates:
            group_to_buttons[group].add(SubstrateButton(substrate,
                                                        (substrate.x, substrate.y),
                                                        (substrate.width, substrate.height)))

    return group_to_buttons


def make_nrps_substrate_buttons():

    name_to_smiles = parse_smiles(PARAS_SMILES)
    proteinogenic_group_to_buttons = make_nrps_substrate_groups(PROTEINOGENIC_SUBSTRATES,
                                                                name_to_smiles)

    nonproteinogenic_group_to_buttons = make_nrps_substrate_groups(NONPROTEINOGENIC_SUBSTRATES,
                                                                   name_to_smiles)
    fattyacid_group_to_buttons = make_nrps_substrate_groups(FATTY_ACIDS,
                                                            name_to_smiles, starter=True)
    nonaminoacid_group_to_buttons = make_nrps_substrate_groups(NON_AMINO_ACIDS,
                                                               name_to_smiles, starter=True)

    return proteinogenic_group_to_buttons, nonproteinogenic_group_to_buttons, fattyacid_group_to_buttons, \
           nonaminoacid_group_to_buttons


def make_fatty_acid_buttons():
    dimensions = (0.2 * WIDTH, HEIGHT / 25)

    fatty_acid_buttons = [FattyAcidSuperOptionButton('Set isoform', (0.25 * WIDTH, 0.52 * HEIGHT), dimensions),
                          FattyAcidSuperOptionButton('Add methyl group', (0.25 * WIDTH, 0.57 * HEIGHT), dimensions),
                          FattyAcidSuperOptionButton('Add amino group', (0.25 * WIDTH, 0.62 * HEIGHT), dimensions),
                          FattyAcidSuperOptionButton('Add -OH group', (0.25 * WIDTH, 0.67 * HEIGHT), dimensions),
                          FattyAcidSuperOptionButton('Add double bond', (0.25 * WIDTH, 0.72 * HEIGHT), dimensions),
                          FattyAcidSuperOptionButton('Set substrate', (0.25 * WIDTH, 0.82 * HEIGHT), dimensions)]

    small_dimensions = (0.1 * WIDTH, HEIGHT / 25)
    options_buttons = [FattyAcidOptionButton('iso', (0.47 * WIDTH, 0.52 * HEIGHT), small_dimensions),
                       FattyAcidOptionButton('anteiso', (0.47 * WIDTH, 0.57 * HEIGHT), dimensions),
                       FattyAcidOptionButton('cis', (0.47 * WIDTH, 0.52 * HEIGHT), small_dimensions),
                       FattyAcidOptionButton('trans', (0.47 * WIDTH, 0.57 * HEIGHT), dimensions),
                       FattyAcidOptionButton('undefined', (0.47 * WIDTH, 0.62 * HEIGHT), dimensions)]

    tiny_dimensions = (0.05 * WIDTH, HEIGHT / 25)

    c_buttons = [CButton(1, (0.47 * WIDTH, 0.52 * HEIGHT), tiny_dimensions),
                 CButton(2, (0.54 * WIDTH, 0.52 * HEIGHT), tiny_dimensions),
                 CButton(3, (0.61 * WIDTH, 0.52 * HEIGHT), tiny_dimensions),
                 CButton(4, (0.68 * WIDTH, 0.52 * HEIGHT), tiny_dimensions),
                 CButton(5, (0.75 * WIDTH, 0.52 * HEIGHT), tiny_dimensions),

                 CButton(6, (0.47 * WIDTH, 0.57 * HEIGHT), tiny_dimensions),
                 CButton(7, (0.54 * WIDTH, 0.57 * HEIGHT), tiny_dimensions),
                 CButton(8, (0.61 * WIDTH, 0.57 * HEIGHT), tiny_dimensions),
                 CButton(9, (0.68 * WIDTH, 0.57 * HEIGHT), tiny_dimensions),
                 CButton(10, (0.75 * WIDTH, 0.57 * HEIGHT), tiny_dimensions),

                 CButton(11, (0.47 * WIDTH, 0.62 * HEIGHT), tiny_dimensions),
                 CButton(12, (0.54 * WIDTH, 0.62 * HEIGHT), tiny_dimensions),
                 CButton(13, (0.61 * WIDTH, 0.62 * HEIGHT), tiny_dimensions),
                 CButton(14, (0.68 * WIDTH, 0.62 * HEIGHT), tiny_dimensions),
                 CButton(15, (0.75 * WIDTH, 0.62 * HEIGHT), tiny_dimensions),

                 CButton(16, (0.47 * WIDTH, 0.67 * HEIGHT), tiny_dimensions),
                 CButton(17, (0.54 * WIDTH, 0.67 * HEIGHT), tiny_dimensions),
                 CButton(18, (0.61 * WIDTH, 0.67 * HEIGHT), tiny_dimensions),
                 CButton(19, (0.68 * WIDTH, 0.67 * HEIGHT), tiny_dimensions),
                 CButton(20, (0.75 * WIDTH, 0.67 * HEIGHT), tiny_dimensions)]

    return fatty_acid_buttons, options_buttons, c_buttons


FATTY_ACID_SUPER_OPTIONS_BUTTONS, FATTY_ACID_OPTIONS_BUTTONS, C_BUTTONS = make_fatty_acid_buttons()

SET_ISOFORM_BUTTON, ADD_METHYL_GROUP_BUTTON, ADD_AMINO_GROUP_BUTTON, ADD_OH_GROUP_BUTTON, \
    ADD_DOUBLE_BOND_BUTTON, SET_SUBSTRATE_BUTTON = FATTY_ACID_SUPER_OPTIONS_BUTTONS

ISO_BUTTON, ANTEISO_BUTTON, CIS_BUTTON, TRANS_BUTTON, UNDEFINED_BUTTON = FATTY_ACID_OPTIONS_BUTTONS

ALL_BUTTONS += FATTY_ACID_SUPER_OPTIONS_BUTTONS
ALL_BUTTONS += FATTY_ACID_OPTIONS_BUTTONS
ALL_BUTTONS += C_BUTTONS


def show_carbon_nr_buttons(screen, active_buttons, mode='size', fatty_acid=None):
    if mode == 'size':
        show_buttons(C_BUTTONS[1:], screen, active_buttons)
    else:
        involved_positions = fatty_acid.count_involvement()
        if mode == 'double bond':
            c_number = fatty_acid.c_nr
            if fatty_acid.isoform:
                c_number -= 1

            double_bond_numbers = [bond[0] for bond in fatty_acid.double_bonds]
            for button in C_BUTTONS[:c_number - 2]:
                double_bond_possible = True
                for nr in double_bond_numbers:
                    if abs(button.nr - nr) <= 1:
                        double_bond_possible = False
                        break

                if button.nr in involved_positions and involved_positions[button.nr] >= 2 or \
                        (button.nr + 1) in involved_positions and involved_positions[button.nr + 1] >= 2:
                    double_bond_possible = False
                if double_bond_possible:
                    show_button(button, screen, active_buttons)

        else:
            c_number = fatty_acid.c_nr
            if fatty_acid.isoform:
                c_number -= 1

            for button in C_BUTTONS[:c_number - 1]:
                modification_possible = True

                if button.nr in involved_positions and involved_positions[button.nr] >= 2:
                    modification_possible = False

                if modification_possible:
                    show_button(button, screen, active_buttons)


def make_pks_substrate_starter_buttons():
    buttons = set()

    name_to_smiles = parse_smiles(PKS_SMILES)

    substrates = []

    for substrate_name, smiles in name_to_smiles.items():
        smiles = name_to_smiles[substrate_name]
        substrate = Substrate(substrate_name, smiles, 'PKS_starter', starter=True)
        substrates.append(substrate)

    substrate_group = SubstrateGroup('pks_starter', substrates)
    for substrate in substrate_group.substrates:
        buttons.add(SubstrateButton(substrate,
                                    (substrate.x, substrate.y),
                                    (substrate.width, substrate.height)))

    return buttons

PKS_STARTER_SUBSTRATE_BUTTONS = make_pks_substrate_starter_buttons()

def make_pks_substrate_buttons():
    buttons = set()

    name_to_smiles = {'malonylcoa': "CC(=O)S",
                      'methylmalonylcoa': "CCC(=O)S",
                      'methoxymalonylcoa': "COCC(=O)S",
                      'ethylmalonylcoa': "CCCC(=O)S",
                      'wildcard': "[*]CC(=O)S"}

    x = {'Malonyl-CoA': 'mal',
         'Methylmalonyl-CoA': 'mmal',
         'Methoxymalonyl-CoA': 'mxmal',
         'Ethylmalonyl-CoA': 'emal',
         'Isobutyryl-CoA': 'isobut',
         '2-Methylbutyryl-CoA': '2metbut',
         'trans-1,2-CPDA': 'trans-1,2-CPDA',
         'Acetyl-CoA': 'Acetyl-CoA',
         'Benzoyl-CoA': 'benz',
         'Propionyl-CoA': 'prop',
         '3-Methylbutyryl-CoA': '3metbut',
         'CE-Malonyl-CoA': 'cemal',
         '2-Rhyd-Malonyl-CoA': '2Rhydmal',
         'CHC-CoA': 'CHC-CoA',
         'inactive': 'inactive'}

    substrates = []

    for substrate_name, smiles in name_to_smiles.items():

        smiles = name_to_smiles[substrate_name]
        substrate = Substrate(substrate_name, smiles, 'PKS')
        substrates.append(substrate)

    substrate_group = SubstrateGroup('pks', substrates)
    for substrate in substrate_group.substrates:
        buttons.add(SubstrateButton(substrate,
                                    (substrate.x, substrate.y),
                                    (substrate.width, substrate.height)))

    return buttons

KR_SUBTYPES = ['', 'A1', 'A2', 'B1', 'B2', 'C1', 'C2']


def make_kr_buttons():
    buttons = set()
    x_coord = int(0.25 * WIDTH)
    y_coord = int(0.52 * HEIGHT)
    for kr_subtype in KR_SUBTYPES:

        buttons.add(KRSubtypeButton((x_coord, y_coord), kr_subtype))
        x_coord += 10 + KR_BUTTON_SIZE

    return buttons


PROTEINOGENIC_BUTTONS, NON_PROTEINOGENIC_BUTTONS, FATTY_ACID_BUTTONS, NON_AMINO_ACID_BUTTONS = \
    make_nrps_substrate_group_buttons()
PKS_SUBSTRATE_BUTTONS = make_pks_substrate_buttons()

PROTEINOGENIC_GROUP_TO_BUTTONS, NONPROTEINOGENIC_GROUP_TO_BUTTONS, FATTY_ACID_GROUP_TO_BUTTONS, \
           NON_AMINOACID_GROUP_TO_BUTTONS = make_nrps_substrate_buttons()

KR_BUTTONS = make_kr_buttons()

ALL_BUTTONS += PROTEINOGENIC_BUTTONS
ALL_BUTTONS += NON_PROTEINOGENIC_BUTTONS
ALL_BUTTONS += FATTY_ACID_BUTTONS
ALL_BUTTONS += NON_AMINO_ACID_BUTTONS
ALL_BUTTONS += PKS_SUBSTRATE_BUTTONS
ALL_BUTTONS += KR_BUTTONS
ALL_BUTTONS += PKS_STARTER_SUBSTRATE_BUTTONS

for buttons in PROTEINOGENIC_GROUP_TO_BUTTONS.values():
    ALL_BUTTONS += buttons

for buttons in NONPROTEINOGENIC_GROUP_TO_BUTTONS.values():
    ALL_BUTTONS += buttons

for buttons in FATTY_ACID_GROUP_TO_BUTTONS.values():
    ALL_BUTTONS += buttons

for buttons in NON_AMINOACID_GROUP_TO_BUTTONS.values():
    ALL_BUTTONS += buttons



import os

from interactive.gene import Gene
from interactive.substrate import Substrate
from interactive.parsers import parse_smiles
import interactive.flatfiles

FLATFILES = os.path.dirname(interactive.flatfiles.__file__)
PARAS_SMILES = os.path.join(FLATFILES, "PARAS_smiles.txt")
PKS_SMILES = os.path.join(FLATFILES, "at_specificities.txt")

pks_name_to_smiles = {'malonylcoa': "CC(=O)S",
                      'methylmalonylcoa': "CCC(=O)S",
                      'methoxymalonylacp': "COCC(=O)S",
                      'ethylmalonylcoa': "CCCC(=O)S",
                      'wildcard': "[*]CC(=O)S"}

nrps_name_to_smiles = parse_smiles(PARAS_SMILES)

pks_starter = parse_smiles(PKS_SMILES)
pks_starter_reverse = {}

for name, smiles in pks_starter.items():
    pks_starter_reverse[smiles] = name


def load_tabular(tabular_file, screen):
    genes = []
    previous_gene_name = None

    with open(tabular_file, 'r') as tabular:
        tabular.readline()
        for line in tabular:
            line_info = line.split('\t')
            for i, item in enumerate(line_info[:]):
                line_info[i] = item.strip()

            if len(line_info) == 5:
                bgc_type, gene_name, module_id, module_type, specificity = line_info
                tailoring = ""
            else:
                bgc_type, gene_name, module_id, module_type, specificity, tailoring = line_info
            tailoring_domains = tailoring.split('|')
            if not specificity:
                specificity = None

            if previous_gene_name != gene_name:
                previous_gene_name = gene_name
                gene = Gene(screen, len(genes))
                gene.name = gene_name
                genes.append(gene)

            module_type = module_type.split('_')[-1].upper()

            genes[-1].add_module(len(genes[-1].modules), module_type)

            for domain in tailoring_domains:
                if domain:
                    subtype = None
                    if domain.startswith('KR'):
                        domain_and_subtype = domain.split('_')
                        if len(domain_and_subtype) == 2:
                            domain, subtype = domain_and_subtype

                    genes[-1].modules[-1].add_domain(domain)
                    if subtype:
                        for dom in genes[-1].modules[-1].domains:
                            if dom.type == 'KR':
                                dom.subtype = subtype

                if specificity:
                    for dom in genes[-1].modules[-1].domains:
                        if dom.type == 'A' or dom.type == 'AT':
                            substrate_smiles = None
                            custom = False
                            if dom.module.type == 'NRPS':
                                if specificity in nrps_name_to_smiles:
                                    substrate_smiles = nrps_name_to_smiles[specificity]
                                    substrate_name = specificity
                                else:
                                    substrate_smiles = specificity
                                    substrate_name = 'FA'
                                    custom = True
                            elif dom.module.type == 'PKS':
                                if specificity in pks_name_to_smiles:
                                    substrate_smiles = pks_name_to_smiles[specificity]
                                    substrate_name = specificity
                                else:
                                    substrate_smiles = specificity
                                    if substrate_smiles in pks_starter_reverse:
                                        substrate_name = pks_starter_reverse[substrate_smiles]
                                    else:
                                        substrate_name = 'start'
                                        custom = True

                            assert substrate_smiles

                            dom.substrate = Substrate(substrate_name, substrate_smiles, dom.module.type, custom=custom)

    return genes









import math
from typing import List
import os

from pikachu.reactions.functional_groups import find_bonds
from pikachu.general import read_smiles
from pikachu.math_functions import Vector

from raichu.data.molecular_moieties import PEPTIDE_BOND
from raichu.data.attributes import AMINOACID_ONE_LETTER_TO_NAME, AMINOACID_ONE_LETTER_TO_SMILES
from raichu.reactions.general_tailoring_reactions import proteolytic_cleavage
from raichu.drawing.drawer import RaichuDrawer
from raichu.attach_to_domain import attach_to_follower_ripp, attach_to_leader_ripp
from raichu.central_chain_detection.label_central_chain import label_nrp_central_chain
from raichu.representations import CleavageSiteRepresentation, MacrocyclizationRepresentation, TailoringRepresentation
from raichu.cluster.base_cluster import Cluster
from raichu.drawing.ripp_drawing import make_circle


class RiPPCluster(Cluster):
    def __init__(self, gene_name_precursor: str, full_peptide: str,
                 core_peptide: str, cleavage_sites: List = None,
                 macrocyclisations: List = None, tailoring_representations=None) -> None:

        super().__init__(tailoring_representations, macrocyclisations)
        self.gene_name_precursor = gene_name_precursor
        self.full_peptide = full_peptide.upper()
        self.core_peptide = core_peptide.upper()
        self.cleavage_sites = cleavage_sites

        self.cleavage_bonds = []

        self.linear_product = None
        self.final_product = None

        self.initialized_macrocyclization_atoms = []
        self.ripp = True

    def write_cluster(self, out_dir):
        if not os.path.exists(out_dir):
            os.mkdir(out_dir)
        cluster_out = os.path.join(out_dir, 'cluster.txt')
        with open(cluster_out, 'w') as out:
            out.write(f"gene_name\t{self.gene_name_precursor}\n")
            out.write(f"full_peptide\t{self.full_peptide}\n")
            out.write(f"core_peptide\t{self.core_peptide}\n")
            cleavage_strings = []
            for cleavage_site in self.cleavage_sites:
                cleavage_strings.append(f"{cleavage_site.amino_acid}|{cleavage_site.amino_acid_index}|{cleavage_site.peptide_fragment_to_keep}")
            cyclic_strings = []
            for cyclic in self.macrocyclisation_representations:
                cyclic_strings.append(f"{cyclic.atom_1}|{cyclic.atom_2}")
            out.write(f"cleavage_sites\t{':'.join(cleavage_strings)}")
            out.write(f"macrocyclisations\t{':'.join(cyclic_strings)}")

        if self.tailoring_representations:
            tailoring_out = os.path.join(out_dir, 'tailoring.txt')
            with open(tailoring_out, 'w') as tailoring:
                tailoring.write('gene_name\ttype\tsubstrate\tmodification_sites\n')

                for enzyme in self.tailoring_representations:

                    if enzyme.modification_sites:
                        site_reprs = []
                        for modification_site in enzyme.modification_sites:

                            site_repr = '|'.join(list(map(str, modification_site)))
                            site_reprs.append(site_repr)

                        site_str = ':'.join(site_reprs)
                    else:
                        site_str = str(None)

                    tailoring.write(f"{enzyme.gene_name}\t{enzyme.type}\t{enzyme.substrate}\t{site_str}\n")

    @classmethod
    def from_file(cls, in_dir):
        in_cluster = os.path.join(in_dir, 'cluster.txt')
        in_tailoring = os.path.join(in_dir, 'tailoring.txt')
        gene_name = None
        full_peptide = None
        core_peptide = None
        cleavage_sites = []
        macrocyclisations = []

        with open(in_cluster, 'r') as cluster:
            for line in cluster:
                line = line.strip()
                if line:
                    category, value = line.split('\t')
                    if category == 'gene_name':
                        gene_name = value
                    elif category == 'full_peptide':
                        full_peptide = value
                    elif category == 'core_peptide':
                        core_peptide = value
                    elif category == 'cleavage_sites':
                        for cleavage_string in value.split(':'):
                            aa, aa_index, fragment_to_keep = cleavage_string.split('|')
                            cleavage_sites.append(CleavageSiteRepresentation(aa, int(aa_index), fragment_to_keep))
                    elif category == 'macrocyclisations':
                        for cyclic_string in value.split(':'):
                            atom_1, atom_2 = cyclic_string.split('|')
                            macrocyclisations.append(MacrocyclizationRepresentation(atom_1, atom_2))

        if os.path.exists(in_tailoring):
            tailoring_enzymes = []
            with open(in_tailoring, 'r') as tailoring:
                tailoring.readline()
                for line in tailoring:
                    line_info = line.split('\t')
                    line_info_cleaned = []
                    for i, entry in enumerate(line_info):
                        if line_info[i] == "None":
                            line_info_cleaned.append(None)
                        else:
                            line_info_cleaned.append(line_info[i])

                    gene_name, enzyme_type, substrate, site_str = line_info_cleaned

                    sites = []

                    if site_str:
                        site = site_str.split(':')
                        sites.append(site.split('|'))
                    tailoring_representation = TailoringRepresentation(gene_name, enzyme_type, sites, substrate)
                    tailoring_enzymes.append(tailoring_representation)

        else:
            tailoring_enzymes = None

        if not macrocyclisations:
            macrocyclisations = None
        if not cleavage_sites:
            cleavage_sites = None

        return cls(gene_name, full_peptide, core_peptide, cleavage_sites, macrocyclisations, tailoring_enzymes)

    def make_peptide(self):
        smiles_peptide_chain = ""
        for amino_acid in self.core_peptide:
            if amino_acid in AMINOACID_ONE_LETTER_TO_SMILES:
                substrate = AMINOACID_ONE_LETTER_TO_SMILES[amino_acid]
            else:
                raise ValueError(f"Unknown amino acid: {amino_acid}")
            smiles_peptide_chain += str(substrate)
        smiles_peptide_chain += "O"
        self.linear_product = read_smiles(smiles_peptide_chain)
        label_nrp_central_chain(
            self.linear_product, module_type='elongation')
        self.chain_intermediate = self.linear_product

    def initialize_cleavage_sites(self) -> None:
        if self.cleavage_sites:
            for cleavage_site in self.cleavage_sites:
                amino_acid_cleavage = cleavage_site.amino_acid
                number_cleavage = cleavage_site.amino_acid_index
                if self.core_peptide[number_cleavage - 1] == amino_acid_cleavage:
                    peptide_bonds = find_bonds(PEPTIDE_BOND, self.linear_product)
                    peptide_bonds = sorted(peptide_bonds, key=lambda bond: bond.nr)
                    cleavage_bond = peptide_bonds[number_cleavage]
                    self.cleavage_bonds += [[cleavage_bond,
                                             cleavage_site.peptide_fragment_to_keep]]
                else:
                    raise ValueError(
                        f"No {AMINOACID_ONE_LETTER_TO_NAME[amino_acid_cleavage]} in position {number_cleavage} for cleavage.")

    def do_proteolytic_cleavage(self):
        self.initialize_cleavage_sites()
        self.cleaved_intermediates.append(self.chain_intermediate.deepcopy())
        for bond, structure_to_keep in self.cleavage_bonds:
            self.chain_intermediate = proteolytic_cleavage(
                bond, self.chain_intermediate, structure_to_keep=structure_to_keep)
            self.cleaved_intermediates.append(self.chain_intermediate.deepcopy())
        self.final_product = self.chain_intermediate

    def draw_precursor(self, fold=10, size=14, as_string=True, out_file=None, amino_acid_sequence=None, leader=True,
                       x_translation=0, y_translation=0, min_padding_y=0):
        if amino_acid_sequence is None:
            amino_acid_sequence = self.full_peptide
        # set begin of chain so that last amino acid is forward
        circles = []
        texts = []
        text_colour = "#000000"
        if leader:

            current_y = min_padding_y + size
            current_x = min([fold, len(amino_acid_sequence)]) * size * 2 + size * 0.5

            forward = False
            # reverse sequence to go backwards
            amino_acid_sequence = amino_acid_sequence[::-1]
            for index, amino_acid in enumerate(amino_acid_sequence):
                # TODO: create dictionary for amino acid colors
                circles.append(make_circle(current_x, current_y, size, amino_acid))
                text = f"""<text x="{current_x}" y="{current_y}" fill="{text_colour}" text-anchor="middle" font-family="verdana" font-size = "{10}">\
                    <tspan y="{current_y}" dy="0.35em">{amino_acid}</tspan></text>"""
                texts.append(text)
                step = size * 2
                if index % fold == 0 and index != 0:
                    forward = not forward
                if (index % fold > fold - 2 or index % fold < 1) and index > 2:
                    step = 2 ** (1 / 2) * step / 2
                    current_y += step
                if forward:
                    current_x += step
                else:
                    current_x -= step

        if not leader:
            current_y = size + y_translation
            current_x = size + x_translation
            forward = True
            for index, amino_acid in enumerate(amino_acid_sequence):
                circles.append(make_circle(current_x, current_y, size, amino_acid))
                text = f"""<text x="{current_x}" y="{current_y}" fill="{text_colour}" text-anchor="middle" font-family="verdana" font-size = "{10}">\
                    <tspan y="{current_y}" dy="0.35em">{amino_acid}</tspan></text>"""
                texts.append(text)
                step = size * 2
                if index % fold == 0 and index != 0:
                    forward = not forward
                if (index % fold > fold - 2 or index % fold < 1) and index > 2:
                    step = 2 ** (1 / 2) * step / 2
                    current_y += step
                if forward:
                    current_x += step
                else:
                    current_x -= step

        svg = ''
        svg += f"""<g id="domain_circles_{leader}">\n"""
        for i, circle in enumerate(circles):
            text = texts[i]
            svg += f"""<g id="domain_bubble_{i}_{leader}">\n"""
            svg += f"{circle}\n"
            svg += f"{text}\n"
            svg += "</g>\n"
        svg += "</g>\n"
        if as_string:
            return svg
        else:
            svg_string = f"""<svg width="{fold * size * 2 + size * 2}" height="{math.ceil(len(amino_acid_sequence) / fold) * size * (2 + 2 ** 0.5) + size}" viewBox="{0} {0} {fold * size * 2 + size} {math.ceil(len(amino_acid_sequence) / fold) * size * 3 + size}" xmlns="http://www.w3.org/2000/svg">\n"""
            svg_string += svg
            svg_string += "</svg>"
            if out_file is None:
                raise ValueError(
                    "Must provide output svg directory if 'as_string' is set to False.")
            else:
                with open(out_file, 'w') as svg_out:
                    svg_out.write(svg_string)

    def draw_cluster(self, fold=10, size=14, as_string=True, out_file=None, add_url=True,
                     draw_cs_in_pink=False):
        leader_pos = Vector(0, 0)
        follower_pos = Vector(0, 0)

        amino_acid_sequence_without_core = self.full_peptide.split(
            self.core_peptide)
        if len(amino_acid_sequence_without_core) != 2:
            raise ValueError("Core peptide not in precursor.")
        [amino_acid_sequence_leader,
         amino_acid_sequence_follower] = amino_acid_sequence_without_core
        structure = self.chain_intermediate.deepcopy()

        if len(amino_acid_sequence_follower) > 0:
            structure = attach_to_follower_ripp(
                structure)
        if len(amino_acid_sequence_leader) > 0:
            structure = attach_to_leader_ripp(
                structure)

        if not self.cyclic_product:

            drawing = RaichuDrawer(structure, dont_show=True, add_url=add_url, ripp=True, horizontal=True,
                                   draw_Cs_in_pink=draw_cs_in_pink)
        else:
            drawing = RaichuDrawer(structure, dont_show=True, add_url=add_url, ripp=True,
                                   make_linear=False, horizontal=True, draw_Cs_in_pink=draw_cs_in_pink)

        drawing.flip_y_axis()
        drawing.move_to_positive_coords()
        drawing.convert_to_int()
        drawing.draw_structure()
        svg_style = drawing.svg_style
        x_translation_leader = 0
        y_translation_leader = 0
        max_x = 0
        max_y = 0
        for atom in drawing.structure.graph:
            if atom.draw.position.x > max_x:
                max_x = atom.draw.position.x
            if atom.draw.position.y > max_y:
                max_y = atom.draw.position.y
            if atom.annotations.domain_type:
                if atom.annotations.domain_type == "Follower":
                    follower_pos = atom.draw.position
                if atom.annotations.domain_type == "Leader":
                    leader_pos = atom.draw.position

        svg_bubbles_leader = self.draw_precursor(fold=fold, size=size, as_string=True,
                                                 amino_acid_sequence=amino_acid_sequence_leader, leader=True,
                                                 min_padding_y=max(leader_pos.y, follower_pos.y))
        position_x_first_bubble_leader = min([fold, len(amino_acid_sequence_leader)]) * size * 2 + size * 0.5
        position_y_first_bubble_leader = max(leader_pos.y, follower_pos.y) + size

        x_translation_leader = position_x_first_bubble_leader - leader_pos.x + 10 - 0.2
        if leader_pos.x != 0:
            y_translation_leader = position_y_first_bubble_leader - leader_pos.y
        else:
            y_translation_leader = position_y_first_bubble_leader - follower_pos.y
        drawing.move_structure(x_translation_leader, y_translation_leader)
        svg = drawing.draw_svg()
        x_translation = position_x_first_bubble_leader - leader_pos.x + 10 - 0.2 + max_x + x_translation_leader
        y_translation = -position_y_first_bubble_leader
        if follower_pos.x != 0:

            y_translation = follower_pos.y - size

            x_translation = follower_pos.x
            svg_bubbles_follower = self.draw_precursor(
                fold=fold, size=size, as_string=True, amino_acid_sequence=amino_acid_sequence_follower, leader=False,
                x_translation=x_translation, y_translation=y_translation)
        else:
            svg_bubbles_follower = ""

        x1 = 0
        x2 = x_translation + 2 * size * min(len(amino_acid_sequence_follower), fold) + 2 * size
        y1 = 0

        y2 = max(max_y + y_translation_leader + 10,
                 (math.ceil(len(amino_acid_sequence_leader) / fold) * size * (1 + 2 ** 0.5) + max_y + 10),
                 y_translation + 2 * size + math.ceil(len(amino_acid_sequence_follower) / fold) * size * (1 + 2 ** 0.5))
        # print(x1, x2, y1, y2)

        svg_string = f"""<svg width="{x2}" height="{y2}" viewBox="{x1} {y1} {x2} {y2}" xmlns="http://www.w3.org/2000/svg">\n"""
        if svg_style:
            svg_string += f"{svg_style}\n"
        svg_string += svg_bubbles_leader
        if follower_pos.x != 0:
            svg_string += svg_bubbles_follower
        svg_string += svg
        svg_string += "</svg>"

        if as_string:
            return svg_string
        else:
            if out_file is None:
                raise ValueError(
                    "Must provide output svg directory if 'as_string' is set to False.")
            else:
                with open(out_file, 'w') as svg_out:
                    svg_out.write(svg_string)


from joblib import dump
import os
from pathlib import Path

from raichu.run_raichu import build_cluster, ClusterRepresentation, ModuleRepresentation, DomainRepresentation
import raichu.test.test_data

# TEST_FOLDER = os.path.dirname(raichu.test.test_data.__file__)
TEST_FOLDER = Path(__file__).resolve().parent

TEST_DATA_FOLDER = TEST_FOLDER / "test_data"


MARFORMYCIN_CLUSTER = ClusterRepresentation([ModuleRepresentation("NRPS", None, "valine",
                                                                  [DomainRepresentation("MfnC", 'A', None, None,
                                                                                        True,
                                                                                        True),
                                                                   DomainRepresentation("MfnC", 'PCP', None, None,
                                                                                        True,
                                                                                        True)
                                                                   ]),
                                             ModuleRepresentation("NRPS", None, "threonine",
                                                                  [DomainRepresentation("MfnC", 'C', None, None,
                                                                                        True,
                                                                                        True),
                                                                   DomainRepresentation("MfnC", 'A', None, None,
                                                                                        True,
                                                                                        True),
                                                                   DomainRepresentation("MfnC", 'PCP', None, None,
                                                                                        True,
                                                                                        True)
                                                                   ]),
                                             ModuleRepresentation("NRPS", None, "tyrosine",
                                                                  [DomainRepresentation("MfnC", 'C', None, None,
                                                                                        True,
                                                                                        True),
                                                                   DomainRepresentation("MfnC", 'A', None, None,
                                                                                        True,
                                                                                        True),
                                                                   DomainRepresentation("MfnC", 'PCP', None, None,
                                                                                        True,
                                                                                        True),

                                                                   DomainRepresentation("MfnC", 'E', None, None,
                                                                                        True,
                                                                                        True),
                                                                   ]),
                                             ModuleRepresentation("NRPS", None, "isoleucine",
                                                                  [DomainRepresentation("MfnD", 'C', None, None,
                                                                                        True,
                                                                                        True),
                                                                   DomainRepresentation("MfnD", 'A', None, None,
                                                                                        True,
                                                                                        True),
                                                                   DomainRepresentation("MfnD", 'PCP', None, None,
                                                                                        True,
                                                                                        True),
                                                                   DomainRepresentation("MfnD", 'E', None, None,
                                                                                        True,
                                                                                        True)
                                                                   ]),
                                             ModuleRepresentation("NRPS", None, "piperazic acid",
                                                                  [DomainRepresentation("MfnE", 'C', None, None,
                                                                                        True,
                                                                                        True),
                                                                   DomainRepresentation("MfnE", 'A', None, None,
                                                                                        True,
                                                                                        True),
                                                                   DomainRepresentation("MfnE", 'PCP', None, None,
                                                                                        True,
                                                                                        True)
                                                                   ]),
                                             ModuleRepresentation("NRPS", None, "leucine",
                                                                  [DomainRepresentation("MfnE", 'A', None, None,
                                                                                        True,
                                                                                        True),
                                                                   DomainRepresentation("MfnE", 'PCP', None, None,
                                                                                        True, False),
                                                                   DomainRepresentation("MfnE", 'C', None, None,
                                                                                        True,
                                                                                        True),
                                                                   DomainRepresentation("MfnE", 'PCP', None, None,
                                                                                        True, True)
                                                                   ]),
                                             ModuleRepresentation("NRPS", None, "valine",
                                                                  [DomainRepresentation("MfnE", 'C', None, None,
                                                                                        True,
                                                                                        True),
                                                                   DomainRepresentation("MfnE", 'A', None, None,
                                                                                        True,
                                                                                        True),
                                                                   DomainRepresentation("MfnE", 'nMT', None, None,
                                                                                        True,
                                                                                        True),
                                                                   DomainRepresentation("MfnE", 'PCP', None, None,
                                                                                        True,
                                                                                        True),
                                                                   DomainRepresentation("MfnE", 'TE', None, None,
                                                                                        True,
                                                                                        True)
                                                                   ])
                                             ]
                                            )


def generate_test_nrps(cluster_representation, cluster_name):
    nrps_folder = os.path.join(TEST_DATA_FOLDER, "nrps")
    if not os.path.exists(nrps_folder):
        os.mkdir(nrps_folder)

    out_folder = os.path.join(nrps_folder, cluster_name)
    if not os.path.exists(out_folder):
        os.mkdir(out_folder)

    cluster = build_cluster(cluster_representation, strict=False)
    cluster.compute_structures(compute_cyclic_products=False)

    for i, intermediate in enumerate(cluster.modular_intermediates):
        dump(intermediate, os.path.join(out_folder, f"module {i + 1}.structure"))


if __name__ == "__main__":
    generate_test_nrps(MARFORMYCIN_CLUSTER, "marformycin")